import * as http from 'http';
import { Server, ProxyServer } from '@refactorjs/http-proxy';
import * as net from 'net';

/**
 * Based on definition by DefinitelyTyped:
 * https://github.com/DefinitelyTyped/DefinitelyTyped/blob/6f529c6c67a447190f86bfbf894d1061e41e07b7/types/http-proxy-middleware/index.d.ts
 */

type NextFunction<T = (err?: any) => void> = T;
interface RequestHandler<TReq = http.IncomingMessage, TRes = http.ServerResponse, TNext = NextFunction> {
    (req: TReq, res: TRes, next?: TNext): void | Promise<void>;
    upgrade: (req: http.IncomingMessage, socket: net.Socket, head: Buffer) => void;
}
type Filter<TReq = http.IncomingMessage> = string | string[] | ((pathname: string, req: TReq) => boolean);
interface Plugin<TReq = http.IncomingMessage, TRes = http.ServerResponse> {
    (proxyServer: ProxyServer, options: Options<TReq, TRes>): void;
}
interface OnProxyEvent<TReq = http.IncomingMessage, TRes = http.ServerResponse> {
    error?: Server.ErrorCallback<Error, TReq, TRes>;
    proxyReq?: Server.ProxyReqCallback<http.ClientRequest, TReq, TRes>;
    proxyReqWs?: Server.ProxyReqWsCallback<http.ClientRequest, TReq>;
    proxyRes?: Server.ProxyResCallback<TReq, TRes>;
    open?: Server.OpenCallback;
    close?: Server.CloseCallback<TReq>;
    start?: Server.StartCallback<TReq, TRes>;
    end?: Server.EndCallback<TReq, TRes>;
    econnreset?: Server.EconnresetCallback<Error, TReq, TRes>;
}
type Logger = Pick<Console, 'info' | 'warn' | 'error'>;
type PathRewrite = {
    [regexp: string]: string;
} | ((path: string, req: http.IncomingMessage) => string | undefined) | ((path: string, req: http.IncomingMessage) => Promise<string>);
interface Options<TReq = http.IncomingMessage, TRes = http.ServerResponse> extends Server.ServerOptions {
    /**
     * Narrow down requests to proxy or not.
     * Filter on {@link http.IncomingMessage.url `pathname`} which is relative to the proxy's "mounting" point in the server.
     * Or use the {@link http.IncomingMessage `req`}  object for more complex filtering.
     * @link https://github.com/chimurai/http-proxy-middleware/blob/master/recipes/pathFilter.md
     * @since v3.0.0
     */
    pathFilter?: Filter<TReq>;
    /**
     * Modify request paths before requests are send to the target.
     * @example
     * ```js
     * createProxyMiddleware({
     *   pathRewrite: {
     *     '^/api/old-path': '/api/new-path', // rewrite path
     *   }
     * });
     * ```
     * @link https://github.com/chimurai/http-proxy-middleware/blob/master/recipes/pathRewrite.md
     */
    pathRewrite?: PathRewrite;
    /**
     * Access the internal http-proxy server instance to customize behavior
     *
     * @example
     * ```js
     * createProxyMiddleware({
     *   plugins: [(proxyServer, options) => {
     *     proxyServer.on('error', (error, req, res) => {
     *       console.error(error);
     *     });
     *   }]
     * });
     * ```
     * @link https://github.com/chimurai/http-proxy-middleware#plugins-array
     * @since v3.0.0
     */
    plugins?: Plugin<TReq, TRes>[];
    /**
     * Eject pre-configured plugins.
     * NOTE: register your own error handlers to prevent server from crashing.
     *
     * @link https://github.com/chimurai/http-proxy-middleware#ejectplugins-boolean-default-false
     * @since v3.0.0
     */
    ejectPlugins?: boolean;
    /**
     * Listen to http-proxy events
     * @see {@link OnProxyEvent} for available events
     * @example
     * ```js
     * createProxyMiddleware({
     *   on: {
     *     error: (error, req, res, target) => {
     *       console.error(error);
     *     }
     *   }
     * });
     * ```
     * @link https://github.com/chimurai/http-proxy-middleware/blob/master/recipes/proxy-events.md
     * @since v3.0.0
     */
    on?: OnProxyEvent<TReq, TRes>;
    /**
     * Dynamically set the {@link Options.target `options.target`}.
     * @example
     * ```js
     * createProxyMiddleware({
     *   router: async (req) => {
     *     return 'http://127:0.0.1:3000';
     *   }
     * });
     * ```
     * @link https://github.com/chimurai/http-proxy-middleware/blob/master/recipes/router.md
     */
    router?: {
        [hostOrPath: string]: Server.ServerOptions['target'];
    } | ((req: TReq) => Server.ServerOptions['target']) | ((req: TReq) => Promise<Server.ServerOptions['target']>);
    /**
     * Log information from http-proxy-middleware
     * @example
     * ```js
     * createProxyMiddleware({
     *  logger: console
     * });
     * ```
     * @link https://github.com/chimurai/http-proxy-middleware/blob/master/recipes/logger.md
     * @since v3.0.0
     */
    logger?: Logger | any;
}

declare class HttpProxyMiddleware<TReq, TRes> {
    #private;
    proxy: ProxyServer;
    constructor(options: Options<TReq, TRes>, proxy?: ProxyServer);
    middleware: RequestHandler;
}

type Interceptor<TReq = http.IncomingMessage, TRes = http.ServerResponse> = (buffer: Buffer, proxyRes: TReq, req: TReq, res: TRes) => Promise<Buffer | string>;
/**
 * Intercept responses from upstream.
 * Automatically decompress (deflate, gzip, brotli).
 * Give developer the opportunity to modify intercepted Buffer and http.ServerResponse
 *
 * NOTE: must set options.selfHandleResponse=true (prevent automatic call of res.end())
 */
declare function responseInterceptor<TReq extends http.IncomingMessage = http.IncomingMessage, TRes extends http.ServerResponse = http.ServerResponse>(interceptor: Interceptor<TReq, TRes>): (proxyRes: TReq, req: TReq, res: TRes) => Promise<void>;

/**
 * Fix proxied body if bodyParser is involved.
 */
declare function fixRequestBody<TReq = http.IncomingMessage>(proxyReq: http.ClientRequest, req: TReq): void;

/**
 * Subscribe to {@link https://www.npmjs.com/package/http-proxy#listening-for-proxy-events http-proxy error events} to prevent server from crashing.
 * Errors are logged with {@link https://www.npmjs.com/package/debug debug} library.
 */
declare const debugProxyErrorsPlugin: Plugin;

declare const errorResponsePlugin: Plugin;

declare const loggerPlugin: Plugin;

/**
 * Implements option.on object to subscribe to http-proxy events.
 *
 * @example
 * ```js
 * createProxyMiddleware({
 *  on: {
 *    error: (error, req, res, target) => {},
 *    proxyReq: (proxyReq, req, res, options) => {},
 *    proxyReqWs: (proxyReq, req, socket, options) => {},
 *    proxyRes: (proxyRes, req, res) => {},
 *    open: (proxySocket) => {},
 *    close: (proxyRes, proxySocket, proxyHead) => {},
 *    start: (req, res, target) => {},
 *    end: (req, res, proxyRes) => {},
 *    econnreset: (error, req, res, target) => {},
 *  }
 * });
 * ```
 */
declare const proxyEventsPlugin: Plugin;

/**
 * @deprecated
 *
 * Will be removed in a future version.
 */
interface LegacyOptions<TReq = http.IncomingMessage, TRes = http.ServerResponse> extends Options<TReq, TRes> {
    /**
     * @deprecated
     * Use `on.error` instead.
     *
     * @example
     * ```js
     * {
     *   on: {
     *    error: () => {}
     * }
     * ```
     */
    onError?: (...args: any[]) => void;
    /**
     * @deprecated
     * Use `on.proxyRes` instead.
     *
     * @example
     * ```js
     * {
     *   on: {
     *    proxyRes: () => {}
     * }
     * ```
     */
    onProxyRes?: (...args: any[]) => void;
    /**
     * @deprecated
     * Use `on.proxyReq` instead.
     *
     * @example
     * ```js
     * {
     *   on: {
     *    proxyReq: () => {}
     * }
     * ```
     */
    onProxyReq?: (...args: any[]) => void;
    /**
     * @deprecated
     * Use `on.proxyReqWs` instead.
     *
     * @example
     * ```js
     * {
     *   on: {
     *    proxyReqWs: () => {}
     * }
     * ```
     */
    onProxyReqWs?: (...args: any[]) => void;
    /**
     * @deprecated
     * Use `on.open` instead.
     *
     * @example
     * ```js
     * {
     *   on: {
     *    open: () => {}
     * }
     * ```
     */
    onOpen?: (...args: any[]) => void;
    /**
     * @deprecated
     * Use `on.close` instead.
     *
     * @example
     * ```js
     * {
     *   on: {
     *    close: () => {}
     * }
     * ```
     */
    onClose?: (...args: any[]) => void;
    /**
     * @deprecated
     * Use `logger` instead.
     *
     * @example
     * ```js
     * {
     *  logger: console
     * }
     * ```
     */
    logProvider?: any;
    /**
     * @deprecated
     * Use `logger` instead.
     *
     * @example
     * ```js
     * {
     *  logger: console
     * }
     * ```
     */
    logLevel?: any;
}

/**
 * @deprecated
 * This function is deprecated and will be removed in a future version.
 *
 * Use {@link createProxyMiddleware} instead.
 */
declare function legacyCreateProxyMiddleware<TReq = http.IncomingMessage, TRes = http.ServerResponse>(shortHand: string): RequestHandler<TReq, TRes>;
declare function legacyCreateProxyMiddleware<TReq = http.IncomingMessage, TRes = http.ServerResponse>(legacyOptions: LegacyOptions<TReq, TRes>): RequestHandler<TReq, TRes>;
declare function legacyCreateProxyMiddleware<TReq = http.IncomingMessage, TRes = http.ServerResponse>(legacyContext: Filter<TReq>, legacyOptions: LegacyOptions<TReq, TRes>): RequestHandler<TReq, TRes>;

declare function createProxyMiddleware<TReq = http.IncomingMessage, TRes = http.ServerResponse, TNext = NextFunction>(options: Options<TReq, TRes>, proxy?: ProxyServer): RequestHandler<TReq, TRes, TNext>;

export { type Filter, HttpProxyMiddleware, type LegacyOptions, type Options, type RequestHandler, createProxyMiddleware, debugProxyErrorsPlugin, errorResponsePlugin, fixRequestBody, legacyCreateProxyMiddleware, loggerPlugin, proxyEventsPlugin, responseInterceptor };
