import { createProxyServer } from '@refactorjs/http-proxy';
import createDebug from 'debug';
import isGlob from 'is-glob';
import micromatch from 'micromatch';
import * as url from 'node:url';
import isPlainObj from 'is-plain-obj';
import * as zlib from 'zlib';
import * as querystring from 'querystring';
import * as url$1 from 'url';

var ERRORS = /* @__PURE__ */ ((ERRORS2) => {
  ERRORS2["ERR_CONFIG_FACTORY_TARGET_MISSING"] = '[HPM] Missing "target" option. Example: {target: "http://www.example.org"}';
  ERRORS2["ERR_CONTEXT_MATCHER_GENERIC"] = '[HPM] Invalid context. Expecting something like: "/api" or ["/api", "/ajax"]';
  ERRORS2["ERR_CONTEXT_MATCHER_INVALID_ARRAY"] = '[HPM] Invalid pathFilter. Expecting something like: ["/api", "/ajax"] or ["/api/**", "!**.html"]';
  ERRORS2["ERR_PATH_REWRITER_CONFIG"] = "[HPM] Invalid pathRewrite config. Expecting object with pathRewrite config or a rewrite function";
  return ERRORS2;
})(ERRORS || {});

function verifyConfig(options) {
  if (!options.target && !options.router) {
    throw new Error(ERRORS.ERR_CONFIG_FACTORY_TARGET_MISSING);
  }
}

const Debug = createDebug("http-proxy-middleware");

const debug$6 = Debug.extend("debug-proxy-errors-plugin");
const debugProxyErrorsPlugin = (proxyServer) => {
  proxyServer.on("error", (error, req, res, target) => {
    debug$6(`http-proxy error event: 
%O`, error);
  });
  proxyServer.on("proxyReq", (proxyReq, req, socket) => {
    socket.on("error", (error) => {
      debug$6("Socket error in proxyReq event: \n%O", error);
    });
  });
  proxyServer.on("proxyRes", (proxyRes, req, res) => {
    res.on("close", () => {
      if (!res.writableEnded) {
        debug$6("Destroying proxyRes in proxyRes close event");
        proxyRes.destroy();
      }
    });
  });
  proxyServer.on("proxyReqWs", (proxyReq, req, socket) => {
    socket.on("error", (error) => {
      debug$6("Socket error in proxyReqWs event: \n%O", error);
    });
  });
  proxyServer.on("open", (proxySocket) => {
    proxySocket.on("error", (error) => {
      debug$6("Socket error in open event: \n%O", error);
    });
  });
  proxyServer.on("close", (req, socket, head) => {
    socket.on("error", (error) => {
      debug$6("Socket error in close event: \n%O", error);
    });
  });
  proxyServer.on("econnreset", (error, req, res, target) => {
    debug$6(`http-proxy econnreset event: 
%O`, error);
  });
};

function getStatusCode(errorCode) {
  let statusCode;
  if (/HPE_INVALID/.test(errorCode)) {
    statusCode = 502;
  } else {
    switch (errorCode) {
      case "ECONNRESET":
      case "ENOTFOUND":
      case "ECONNREFUSED":
      case "ETIMEDOUT":
        statusCode = 504;
        break;
      default:
        statusCode = 500;
        break;
    }
  }
  return statusCode;
}

const errorResponsePlugin = (proxyServer, options) => {
  proxyServer.on("error", (err, req, res, target) => {
    if (!req && !res) {
      throw err;
    }
    if ("writeHead" in res && !res.headersSent) {
      const statusCode = getStatusCode(err.code);
      res.writeHead(statusCode);
    }
    const host = req.headers && req.headers.host;
    res.end(`Error occurred while trying to proxy: ${host}${req.url}`);
  });
};

const noopLogger = {
  info: () => {
  },
  warn: () => {
  },
  error: () => {
  }
};
function getLogger(options) {
  return options.logger || noopLogger;
}

const loggerPlugin = (proxyServer, options) => {
  const logger = getLogger(options);
  proxyServer.on("error", (err, req, res, target) => {
    const hostname = req?.headers?.host;
    const requestHref = `${hostname}${req?.url}`;
    const targetHref = `${target?.href}`;
    const errorMessage = "[HPM] Error occurred while proxying request %s to %s [%s] (%s)";
    const errReference = "https://nodejs.org/api/errors.html#errors_common_system_errors";
    logger.error(errorMessage, requestHref, targetHref, err.code || err, errReference);
  });
  proxyServer.on("proxyRes", (proxyRes, req, res) => {
    const originalUrl = req.originalUrl ?? `${req.baseUrl || ""}${req.url}`;
    const exchange = `[HPM] ${req.method} ${originalUrl} -> ${proxyRes.req.protocol}//${proxyRes.req.host}${proxyRes.req.path} [${proxyRes.statusCode}]`;
    logger.info(exchange);
  });
  proxyServer.on("open", (socket) => {
    logger.info("[HPM] Client connected: %o", socket.address());
  });
  proxyServer.on("close", (req, proxySocket, proxyHead) => {
    logger.info("[HPM] Client disconnected: %o", proxySocket.address());
  });
};

function getFunctionName(fn) {
  return fn.name || "[anonymous Function]";
}

const debug$5 = Debug.extend("proxy-events-plugin");
const proxyEventsPlugin = (proxyServer, options) => {
  Object.entries(options.on || {}).forEach(([eventName, handler]) => {
    debug$5(`register event handler: "${eventName}" -> "${getFunctionName(handler)}"`);
    proxyServer.on(eventName, handler);
  });
};

function getPlugins(options) {
  const maybeErrorResponsePlugin = !!options.on?.error ? [] : [errorResponsePlugin];
  const defaultPlugins = !!options.ejectPlugins ? [] : [debugProxyErrorsPlugin, proxyEventsPlugin, loggerPlugin, ...maybeErrorResponsePlugin];
  const userPlugins = options.plugins ?? [];
  return [...defaultPlugins, ...userPlugins];
}

function matchPathFilter(pathFilter = "/", uri, req) {
  if (isStringPath(pathFilter)) {
    return matchSingleStringPath(pathFilter, uri);
  }
  if (isGlobPath(pathFilter)) {
    return matchSingleGlobPath(pathFilter, uri);
  }
  if (Array.isArray(pathFilter)) {
    if (pathFilter.every(isStringPath)) {
      return matchMultiPath(pathFilter, uri);
    }
    if (pathFilter.every(isGlobPath)) {
      return matchMultiGlobPath(pathFilter, uri);
    }
    throw new Error(ERRORS.ERR_CONTEXT_MATCHER_INVALID_ARRAY);
  }
  if (typeof pathFilter === "function") {
    const pathname = getUrlPathName(uri);
    return pathFilter(pathname, req);
  }
  throw new Error(ERRORS.ERR_CONTEXT_MATCHER_GENERIC);
}
function matchSingleStringPath(pathFilter, uri) {
  const pathname = getUrlPathName(uri);
  return pathname?.indexOf(pathFilter) === 0;
}
function matchSingleGlobPath(pattern, uri) {
  const pathname = getUrlPathName(uri);
  const matches = micromatch([pathname], pattern);
  return matches && matches.length > 0;
}
function matchMultiGlobPath(patternList, uri) {
  return matchSingleGlobPath(patternList, uri);
}
function matchMultiPath(pathFilterList, uri) {
  let isMultiPath = false;
  for (const context of pathFilterList) {
    if (matchSingleStringPath(context, uri)) {
      isMultiPath = true;
      break;
    }
  }
  return isMultiPath;
}
function getUrlPathName(uri) {
  return uri && url.parse(uri).pathname;
}
function isStringPath(pathFilter) {
  return typeof pathFilter === "string" && !isGlob(pathFilter);
}
function isGlobPath(pathFilter) {
  return isGlob(pathFilter);
}

const debug$4 = Debug.extend("path-rewriter");
function createPathRewriter(rewriteConfig) {
  let rulesCache;
  if (!isValidRewriteConfig(rewriteConfig)) {
    return;
  }
  if (typeof rewriteConfig === "function") {
    const customRewriteFn = rewriteConfig;
    return customRewriteFn;
  } else {
    rulesCache = parsePathRewriteRules(rewriteConfig);
    return rewritePath;
  }
  function rewritePath(path) {
    let result = path;
    for (const rule of rulesCache) {
      if (rule.regex.test(path)) {
        result = result.replace(rule.regex, rule.value);
        debug$4('rewriting path from "%s" to "%s"', path, result);
        break;
      }
    }
    return result;
  }
}
function isValidRewriteConfig(rewriteConfig) {
  if (typeof rewriteConfig === "function") {
    return true;
  } else if (isPlainObj(rewriteConfig)) {
    return Object.keys(rewriteConfig).length !== 0;
  } else if (rewriteConfig === void 0 || rewriteConfig === null) {
    return false;
  } else {
    throw new Error(ERRORS.ERR_PATH_REWRITER_CONFIG);
  }
}
function parsePathRewriteRules(rewriteConfig) {
  const rules = [];
  if (isPlainObj(rewriteConfig)) {
    for (const [key, value] of Object.entries(rewriteConfig)) {
      rules.push({
        regex: new RegExp(key),
        value
      });
      debug$4('rewrite rule created: "%s" ~> "%s"', key, value);
    }
  }
  return rules;
}

const debug$3 = Debug.extend("router");
async function getTarget(req, config) {
  let newTarget;
  const router = config.router;
  if (isPlainObj(router)) {
    newTarget = getTargetFromProxyTable(req, router);
  } else if (typeof router === "function") {
    newTarget = await router(req);
  }
  return newTarget;
}
function getTargetFromProxyTable(req, table) {
  let result;
  const host = req.headers.host;
  const path = req.url;
  const hostAndPath = host + path;
  for (const [key, value] of Object.entries(table)) {
    if (containsPath(key)) {
      if (hostAndPath.indexOf(key) > -1) {
        result = value;
        debug$3('match: "%s" -> "%s"', key, result);
        break;
      }
    } else {
      if (key === host) {
        result = value;
        debug$3('match: "%s" -> "%s"', host, result);
        break;
      }
    }
  }
  return result;
}
function containsPath(v) {
  return v.indexOf("/") > -1;
}

var __defProp = Object.defineProperty;
var __defNormalProp = (obj, key, value) => key in obj ? __defProp(obj, key, { enumerable: true, configurable: true, writable: true, value }) : obj[key] = value;
var __publicField = (obj, key, value) => {
  __defNormalProp(obj, typeof key !== "symbol" ? key + "" : key, value);
  return value;
};
var __accessCheck = (obj, member, msg) => {
  if (!member.has(obj))
    throw TypeError("Cannot " + msg);
};
var __privateGet = (obj, member, getter) => {
  __accessCheck(obj, member, "read from private field");
  return getter ? getter.call(obj) : member.get(obj);
};
var __privateAdd = (obj, member, value) => {
  if (member.has(obj))
    throw TypeError("Cannot add the same private member more than once");
  member instanceof WeakSet ? member.add(obj) : member.set(obj, value);
};
var __privateSet = (obj, member, value, setter) => {
  __accessCheck(obj, member, "write to private field");
  setter ? setter.call(obj, value) : member.set(obj, value);
  return value;
};
var __privateMethod = (obj, member, method) => {
  __accessCheck(obj, member, "access private method");
  return method;
};
var _wsInternalSubscribed, _serverOnCloseSubscribed, _proxyOptions, _pathRewriter, _registerPlugins, registerPlugins_fn, _catchUpgradeRequest, _handleUpgrade, _shouldProxy, _prepareProxyRequest, _applyRouter, _applyPathRewrite;
class HttpProxyMiddleware {
  constructor(options, proxy) {
    __privateAdd(this, _registerPlugins);
    __privateAdd(this, _wsInternalSubscribed, false);
    __privateAdd(this, _serverOnCloseSubscribed, false);
    __privateAdd(this, _proxyOptions, void 0);
    __privateAdd(this, _pathRewriter, void 0);
    __publicField(this, "proxy");
    // https://github.com/Microsoft/TypeScript/wiki/'this'-in-TypeScript#red-flags-for-this
    __publicField(this, "middleware", async (req, res, next) => {
      if (__privateGet(this, _shouldProxy).call(this, __privateGet(this, _proxyOptions).pathFilter, req)) {
        try {
          const activeProxyOptions = await __privateGet(this, _prepareProxyRequest).call(this, req);
          Debug(`proxy request to target: %O`, activeProxyOptions.target);
          this.proxy.web(req, res, activeProxyOptions);
        } catch (err) {
          next && next(err);
        }
      } else {
        next && next();
      }
      const server = (req.socket ?? req.connection)?.server;
      if (server && !__privateGet(this, _serverOnCloseSubscribed) && __privateGet(this, _proxyOptions).ws === true) {
        server.on("close", () => {
          Debug("server close signal received: closing proxy server");
          this.proxy.close();
        });
        __privateSet(this, _serverOnCloseSubscribed, true);
      }
      if (__privateGet(this, _proxyOptions).ws === true) {
        __privateGet(this, _catchUpgradeRequest).call(this, server);
      }
    });
    __privateAdd(this, _catchUpgradeRequest, (server) => {
      if (!__privateGet(this, _wsInternalSubscribed)) {
        Debug("subscribing to server upgrade event");
        server.on("upgrade", __privateGet(this, _handleUpgrade));
        __privateSet(this, _wsInternalSubscribed, true);
      }
    });
    __privateAdd(this, _handleUpgrade, async (req, socket, head) => {
      if (__privateGet(this, _shouldProxy).call(this, __privateGet(this, _proxyOptions).pathFilter, req)) {
        const activeProxyOptions = await __privateGet(this, _prepareProxyRequest).call(this, req);
        this.proxy.ws(req, socket, head, activeProxyOptions);
        Debug("server upgrade event received. Proxying WebSocket");
      }
    });
    /**
     * Determine whether request should be proxied.
     */
    __privateAdd(this, _shouldProxy, (pathFilter, req) => {
      return matchPathFilter(pathFilter, req.url, req);
    });
    /**
     * Apply option.router and option.pathRewrite
     * Order matters:
     *    Router uses original path for routing;
     *    NOT the modified path, after it has been rewritten by pathRewrite
     * @param {Object} req
     * @return {Object} proxy options
     */
    __privateAdd(this, _prepareProxyRequest, async (req) => {
      if (this.middleware.__LEGACY_HTTP_PROXY_MIDDLEWARE__) {
        req.url = req.originalUrl || req.url;
      }
      const newProxyOptions = Object.assign({}, __privateGet(this, _proxyOptions));
      await __privateGet(this, _applyRouter).call(this, req, newProxyOptions);
      await __privateGet(this, _applyPathRewrite).call(this, req, __privateGet(this, _pathRewriter));
      return newProxyOptions;
    });
    // Modify option.target when router present.
    __privateAdd(this, _applyRouter, async (req, options) => {
      let newTarget;
      if (options.router) {
        newTarget = await getTarget(req, options);
        if (newTarget) {
          Debug('router new target: "%s"', newTarget);
          options.target = newTarget;
        }
      }
    });
    // rewrite path
    __privateAdd(this, _applyPathRewrite, async (req, pathRewriter) => {
      if (pathRewriter) {
        const path = await pathRewriter(req.url, req);
        if (typeof path === "string") {
          Debug("pathRewrite new path: %s", req.url);
          req.url = path;
        } else {
          Debug("pathRewrite: no rewritten path found: %s", req.url);
        }
      }
    });
    verifyConfig(options);
    __privateSet(this, _proxyOptions, options);
    Debug(`create proxy server`);
    this.proxy = proxy ?? createProxyServer({});
    __privateMethod(this, _registerPlugins, registerPlugins_fn).call(this, this.proxy, __privateGet(this, _proxyOptions));
    __privateSet(this, _pathRewriter, createPathRewriter(__privateGet(this, _proxyOptions).pathRewrite));
    this.middleware.upgrade = (req, socket, head) => {
      if (!__privateGet(this, _wsInternalSubscribed)) {
        __privateGet(this, _handleUpgrade).call(this, req, socket, head);
      }
    };
  }
}
_wsInternalSubscribed = new WeakMap();
_serverOnCloseSubscribed = new WeakMap();
_proxyOptions = new WeakMap();
_pathRewriter = new WeakMap();
_registerPlugins = new WeakSet();
registerPlugins_fn = function(proxy, options) {
  const plugins = getPlugins(options);
  plugins.forEach((plugin) => {
    Debug(`register plugin: "${getFunctionName(plugin)}"`);
    plugin(proxy, options);
  });
};
_catchUpgradeRequest = new WeakMap();
_handleUpgrade = new WeakMap();
_shouldProxy = new WeakMap();
_prepareProxyRequest = new WeakMap();
_applyRouter = new WeakMap();
_applyPathRewrite = new WeakMap();

const debug$2 = Debug.extend("response-interceptor");
const TrailerDisallowHeaders = [
  "content-length",
  "host",
  "content-type",
  "authorization",
  "cache-control",
  "max-forwards",
  "te",
  "set-cookie",
  "content-encoding",
  "content-range"
];
function responseInterceptor(interceptor) {
  return async function proxyResResponseInterceptor(proxyRes, req, res) {
    debug$2("intercept proxy response");
    const originalProxyRes = proxyRes;
    let buffer = Buffer.from("", "utf8");
    const _proxyRes = decompress(proxyRes, proxyRes.headers["content-encoding"]);
    _proxyRes.on("data", (chunk) => buffer = Buffer.concat([buffer, chunk]));
    _proxyRes.on("end", async () => {
      copyHeaders(proxyRes, res);
      debug$2("call interceptor function: %s", getFunctionName(interceptor));
      const interceptedBuffer = Buffer.from(await interceptor(buffer, originalProxyRes, req, res));
      debug$2("set content-length: %s", Buffer.byteLength(interceptedBuffer, "utf8"));
      if (proxyRes.headers.trailer === void 0) {
        res.setHeader("content-length", Buffer.byteLength(interceptedBuffer, "utf8"));
      } else {
        TrailerDisallowHeaders.forEach((value) => {
          res.removeHeader(value);
        });
      }
      res.setHeader("content-length", Buffer.byteLength(interceptedBuffer, "utf8"));
      debug$2("write intercepted response");
      res.write(interceptedBuffer);
      res.end();
    });
    _proxyRes.on("error", (error) => {
      res.end(`Error fetching proxied request: ${error.message}`);
    });
  };
}
function decompress(proxyRes, contentEncoding) {
  let _proxyRes = proxyRes;
  let decompress2;
  switch (contentEncoding) {
    case "gzip":
      decompress2 = zlib.createGunzip();
      break;
    case "br":
      decompress2 = zlib.createBrotliDecompress();
      break;
    case "deflate":
      decompress2 = zlib.createInflate();
      break;
  }
  if (decompress2) {
    debug$2(`decompress proxy response with 'content-encoding': %s`, contentEncoding);
    _proxyRes.pipe(decompress2);
    _proxyRes = decompress2;
  }
  return _proxyRes;
}
function copyHeaders(originalResponse, response) {
  debug$2("copy original response headers");
  response.statusCode = originalResponse.statusCode;
  response.statusMessage = originalResponse.statusMessage;
  if (response.setHeader) {
    let keys = Object.keys(originalResponse.headers);
    keys = keys.filter((key) => !["content-encoding", "transfer-encoding"].includes(key));
    keys.forEach((key) => {
      let value = originalResponse.headers[key];
      if (key === "set-cookie") {
        value = Array.isArray(value) ? value : [value];
        value = value.map((x) => x.replace(/Domain=[^;]+?/i, ""));
      }
      response.setHeader(key, value);
    });
  } else {
    response.headers = originalResponse.headers;
  }
}

function fixRequestBody(proxyReq, req) {
  const requestBody = req.body;
  if (!requestBody) {
    return;
  }
  const contentType = proxyReq.getHeader("Content-Type");
  const writeBody = (bodyData) => {
    proxyReq.setHeader("Content-Length", Buffer.byteLength(bodyData));
    proxyReq.write(bodyData);
  };
  if (contentType && contentType.includes("application/json")) {
    writeBody(JSON.stringify(requestBody));
  }
  if (contentType && contentType.includes("application/x-www-form-urlencoded")) {
    writeBody(querystring.stringify(requestBody));
  }
  if (contentType && contentType.includes("multipart/form-data")) {
    writeBody(handlerFormDataBodyData(contentType, requestBody));
  }
}
function handlerFormDataBodyData(contentType, data) {
  const boundary = contentType.replace(/^.*boundary=(.*)$/, "$1");
  let str = "";
  for (const [key, value] of Object.entries(data)) {
    str += `--${boundary}\r
Content-Disposition: form-data; name="${key}"\r
\r
${value}\r
`;
  }
  return str;
}

const debug$1 = Debug.extend("legacy-options-adapter");
const proxyEventMap = {
  onError: "error",
  onProxyReq: "proxyReq",
  onProxyRes: "proxyRes",
  onProxyReqWs: "proxyReqWs",
  onOpen: "open",
  onClose: "close"
};
function legacyOptionsAdapter(legacyContext, legacyOptions) {
  let options = {};
  let logger;
  if (typeof legacyContext === "string" && !!url$1.parse(legacyContext).host) {
    throw new Error(
      `Shorthand syntax is removed from legacyCreateProxyMiddleware().
      Please use "legacyCreateProxyMiddleware({ target: 'http://www.example.org' })" instead.

      More details: https://github.com/chimurai/http-proxy-middleware/blob/master/MIGRATION.md#removed-shorthand-usage
      `
    );
  }
  if (legacyContext && legacyOptions) {
    debug$1("map legacy context/filter to options.pathFilter");
    options = { ...legacyOptions, pathFilter: legacyContext };
    logger = getLegacyLogger(options);
    logger.warn(
      `[http-proxy-middleware] Legacy "context" argument is deprecated. Migrate your "context" to "options.pathFilter":

      const options = {
        pathFilter: '${legacyContext}',
      }

      More details: https://github.com/chimurai/http-proxy-middleware/blob/master/MIGRATION.md#removed-context-argument
      `
    );
  } else if (legacyContext && !legacyOptions) {
    options = { ...legacyContext };
    logger = getLegacyLogger(options);
  } else {
    logger = getLegacyLogger({});
  }
  Object.entries(proxyEventMap).forEach(([legacyEventName, proxyEventName]) => {
    if (options[legacyEventName]) {
      options.on = { ...options.on };
      options.on[proxyEventName] = options[legacyEventName];
      debug$1('map legacy event "%s" to "on.%s"', legacyEventName, proxyEventName);
      logger.warn(
        `[http-proxy-middleware] Legacy "${legacyEventName}" is deprecated. Migrate to "options.on.${proxyEventName}":

        const options = {
          on: {
            ${proxyEventName}: () => {},
          },
        }

        More details: https://github.com/chimurai/http-proxy-middleware/blob/master/MIGRATION.md#refactored-proxy-events
        `
      );
    }
  });
  const logProvider = options.logProvider && options.logProvider();
  const logLevel = options.logLevel;
  debug$1("legacy logLevel", logLevel);
  debug$1("legacy logProvider: %O", logProvider);
  if (typeof logLevel === "string" && logLevel !== "silent") {
    debug$1('map "logProvider" to "logger"');
    logger.warn(
      `[http-proxy-middleware] Legacy "logLevel" and "logProvider" are deprecated. Migrate to "options.logger":

      const options = {
        logger: console,
      }

      More details: https://github.com/chimurai/http-proxy-middleware/blob/master/MIGRATION.md#removed-logprovider-and-loglevel-options
      `
    );
  }
  return options;
}
function getLegacyLogger(options) {
  const legacyLogger = options.logProvider && options.logProvider();
  if (legacyLogger) {
    options.logger = legacyLogger;
  }
  return getLogger(options);
}

const debug = Debug.extend("legacy-create-proxy-middleware");
function legacyCreateProxyMiddleware(legacyContext, legacyOptions) {
  debug("init");
  const options = legacyOptionsAdapter(legacyContext, legacyOptions);
  const proxyMiddleware = createProxyMiddleware(options);
  debug("add marker for patching req.url (old behavior)");
  proxyMiddleware.__LEGACY_HTTP_PROXY_MIDDLEWARE__ = true;
  return proxyMiddleware;
}

function createProxyMiddleware(options, proxy) {
  const { middleware } = new HttpProxyMiddleware(options, proxy);
  return middleware;
}

export { HttpProxyMiddleware, createProxyMiddleware, debugProxyErrorsPlugin, errorResponsePlugin, fixRequestBody, legacyCreateProxyMiddleware, loggerPlugin, proxyEventsPlugin, responseInterceptor };
